import logging
from threading import Thread, Event
from queue import Queue, Empty
import time
import pika
import pika.exceptions
import wagglemsg
from .config import PluginConfig


logger = logging.getLogger(__name__)
# pika is very verbose at DEBUG level. we turn it down here.
logging.getLogger("pika").setLevel(logging.CRITICAL)


class RabbitMQPublisher:
    """
    RabbitMQPublisher manages a connection to RabbitMQ and publishes messages from the provided queue.

    This is done in a background thread which must be stopped by setting the provided stop Event.
    """

    def __init__(self, config: PluginConfig, messages: Queue, stop: Event):
        self.config = config
        self.params = get_connection_parameters_for_config(config)
        self.messages = messages
        self.stop = stop
        self.done = Event()
        Thread(target=self.__main).start()

    def __main(self):
        logger.debug("publisher started.")
        try:
            while not self.stop.is_set():
                try:
                    self.__connect_and_flush_messages()
                except Exception:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception("__connect_and_flush_messages exception")
                    time.sleep(1)
        finally:
            self.done.set()
            logger.debug("publisher stopped.")

    def __connect_and_flush_messages(self):
        logger.debug("publisher connecting to rabbitmq...")
        with pika.BlockingConnection(self.params) as conn, conn.channel() as ch:
            while not self.stop.is_set():
                self.__flush_messages(ch)
            logger.debug("publisher stopping...")
            # attempt to flush any remaining messages
            self.__flush_messages(ch)

    def __flush_messages(self, ch):
        while True:
            try:
                logger.debug("publisher checking for message...")
                item = self.messages.get(timeout=1)
            except Empty:
                return

            properties = pika.BasicProperties(
                delivery_mode=2, user_id=self.params.credentials.username
            )

            # NOTE app_id is used by data service to validate and tag additional metadata provided by k3s scheduler.
            if self.config.app_id != "":
                properties.app_id = self.config.app_id

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(
                    "publishing message to rabbitmq: %s", wagglemsg.load(item.body)
                )

            try:
                ch.basic_publish(
                    exchange="to-validator",
                    routing_key=item.scope,
                    properties=properties,
                    body=item.body,
                )
            except Exception:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception(
                        "basic_publish to rabbitmq failed. will requeue message..."
                    )
                # requeue message so we can again later
                # NOTE(sean) this will reorder messages. if we realized we *must* preserve message
                # order, we must to change this to avoid subtle bugs!
                self.messages.put(item)
                # propagate error up to trigger reconnect
                raise


class RabbitMQConsumer:
    """
    RabbitMQConsumer manages a connection to RabbitMQ and puts received messages into the provided queue.

    This is done in a background thread which must be stopped by setting the provided stop Event.
    """

    def __init__(self, topics, config: PluginConfig, messages: Queue, stop: Event):
        self.topics = topics
        self.config = config
        self.params = get_connection_parameters_for_config(config)
        self.messages = messages
        self.stop = stop
        self.done = Event()
        Thread(target=self.__main).start()

    def __main(self):
        logger.debug("consumer started.")
        try:
            while not self.stop.is_set():
                try:
                    self.__connect_and_consume_messages()
                except Exception:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.exception("__connect_and_consume_messages exception")
                    time.sleep(1)
        finally:
            self.done.set()
            logger.debug("consumer stopped.")

    def __connect_and_consume_messages(self):
        logger.debug("consumer connecting to rabbitmq...")
        with pika.BlockingConnection(self.params) as conn, conn.channel() as ch:
            # setup subscriber queue and bind to topics
            queue = ch.queue_declare("", exclusive=True).method.queue
            ch.basic_consume(queue, self.__process_message, auto_ack=True)

            for topic in self.topics:
                ch.queue_bind(queue, "data.topic", topic)
                logger.debug("consumer binding queue %s to topic %s", queue, topic)

            def check_stop():
                if self.stop.is_set():
                    logger.debug("consumer stopping...")
                    ch.stop_consuming()
                else:
                    conn.call_later(1, check_stop)

            conn.call_later(1, check_stop)
            logger.debug("consumer start processing messages...")
            ch.start_consuming()

    def __process_message(self, ch, method, properties, body):
        try:
            logger.debug("consumer processing message %s...", body)
            msg = wagglemsg.load(body)
        except TypeError:
            logger.debug("unsupported message type: %s %s", properties, body)
            return
        logger.debug("consumer putting message in waiting queue")
        self.messages.put(msg)


def get_connection_parameters_for_config(
    config: PluginConfig,
) -> pika.ConnectionParameters:
    return pika.ConnectionParameters(
        host=config.host,
        port=config.port,
        credentials=pika.PlainCredentials(
            username=config.username,
            password=config.password,
        ),
        connection_attempts=1,
        socket_timeout=1.0,
    )



import unittest
from pathlib import Path
import json
from tempfile import TemporaryDirectory
import time
from datetime import datetime
import os
import pika
import subprocess

from waggle.plugin import Plugin, PluginConfig, Uploader, get_timestamp
import wagglemsg

# TODO(sean) add integration testing against rabbitmq
# TODO(sean) clean up the queue interface. it would be better to not know about the plugin.send / plugin.recv queues explicitly.


class TestPlugin(unittest.TestCase):
    def test_publish(self):
        with Plugin() as plugin:
            plugin.publish("test.int", 1)
            plugin.publish("test.float", 2.0)
            plugin.publish("test.str", "three")
            plugin.publish(
                "cows.total",
                391,
                meta={
                    "camera": "bottom_left",
                },
            )

    def test_publish_check_reserved(self):
        with Plugin() as plugin:
            with self.assertRaises(ValueError):
                plugin.publish("upload", "path/to/data")

    def test_get(self):
        with Plugin() as plugin:
            plugin.subscribe("raw.#")
            with self.assertRaises(TimeoutError):
                plugin.get(timeout=0)
            with self.assertRaises(TimeoutError):
                plugin.get(timeout=0.001)

            msg = wagglemsg.Message("test", 1.0, 0, {})
            plugin.recv.put(msg)
            msg2 = plugin.get(timeout=0)
            self.assertEqual(msg, msg2)

    def test_get_timestamp(self):
        ts = get_timestamp()
        self.assertIsInstance(ts, int)

    def test_valid_values(self):
        with Plugin() as plugin:
            plugin.publish("test", 1)
            plugin.publish("test", 1.3)
            plugin.publish("test", "some string")
            with self.assertRaises(TypeError):
                plugin.publish("test", b"some bytes")
            with self.assertRaises(TypeError):
                plugin.publish("test", [1, 2, 3])
            with self.assertRaises(TypeError):
                plugin.publish("test", {1: 1, 2: 2, 3: 3})

    def test_valid_meta(self):
        with Plugin() as plugin:
            plugin.publish("test", 1, meta={"k": "v"})
            with self.assertRaises(TypeError):
                plugin.publish("test", 1, meta={"k": 10})
            with self.assertRaises(TypeError):
                plugin.publish("test", 1, meta={"k": 12.3})
            with self.assertRaises(TypeError):
                plugin.publish("test", 1, meta={"k": []})

    def test_valid_timestamp(self):
        with Plugin() as plugin:
            # valid int, nanosecond timestamp
            plugin.publish("test", 1, timestamp=1649694687904754000)

            # must prevent a float type timestamp
            ts = datetime(2022, 1, 1, 0, 0, 0).timestamp()
            with self.assertRaises(TypeError):
                plugin.publish("test", 1, timestamp=ts)

            # must prevent int timestamp in seconds from being loaded by flagging
            # timestamps that are too early.
            testcases = [
                datetime(2022, 1, 1, 0, 0, 0),
                datetime(3000, 1, 1, 0, 0, 0),
                datetime(5000, 1, 1, 0, 0, 0),
            ]

            for dt in testcases:
                with self.assertRaises(ValueError):
                    plugin.publish("test", 1, timestamp=int(dt.timestamp()))

    def test_valid_publish_names(self):
        with Plugin() as plugin:
            with self.assertRaises(TypeError):
                plugin.publish(None, 0)
            with self.assertRaises(ValueError):
                plugin.publish("", 0)
            with self.assertRaises(ValueError):
                plugin.publish(".", 0)

            # check for reserved names
            with self.assertRaises(ValueError):
                plugin.publish("upload", 0)

            # use _ instead of -
            with self.assertRaises(ValueError):
                plugin.publish("my-metric", 0)
            # correct alternative
            plugin.publish("my_metric", 0)

            # assert len(name) <= 128
            with self.assertRaises(ValueError):
                plugin.publish("x" * 129, 0)

            # no empty parts allowed
            with self.assertRaises(ValueError):
                plugin.publish("vision.count..bird", 0)
            # correct alternative
            plugin.publish("vision.count.bird", 0)

            # no spaces allowed
            with self.assertRaises(ValueError):
                plugin.publish("sys.cpu temp", 0)
            # correct alternative
            plugin.publish("sys.cpu_temp", 0)

    # TODO(sean) refactor messaging part to make testing this cleaner
    def test_upload_file(self):
        with TemporaryDirectory() as tempdir:
            pl = Plugin(
                PluginConfig(
                    host="fake-rabbitmq-host",
                    port=5672,
                    username="plugin",
                    password="plugin",
                    app_id="0668b12c-0c15-462c-9e06-7239282411e5",
                ),
                uploader=Uploader(Path(tempdir, "uploads")),
            )

            data = b"here some data in a data"
            upload_path = Path(tempdir, "myfile.txt")
            upload_path.write_bytes(data)

            pl.upload_file(upload_path)
            item = pl.send.get_nowait()
            msg = wagglemsg.load(item.body)
            self.assertEqual(item.scope, "all")
            self.assertEqual(msg.name, "upload")
            self.assertIsNotNone(msg.timestamp)
            self.assertIsInstance(msg.value, str)
            self.assertIsNotNone(msg.meta)
            self.assertIn("filename", msg.meta)

    def test_timeit(self):
        with Plugin() as plugin:
            with plugin.timeit("dur"):
                time.sleep(0.001)
            item = plugin.send.get(0.01)
            msg = wagglemsg.load(item.body)
            self.assertEqual(msg.name, "dur")


class TestUploader(unittest.TestCase):
    def test_upload_file(self):
        with TemporaryDirectory() as tempdir:
            uploader = Uploader(Path(tempdir, "uploads"))

            data = b"here some data in a data"

            upload_path = Path(tempdir, "myfile.txt")
            upload_path.write_bytes(data)

            path = uploader.upload_file(upload_path)
            self.assertFalse(upload_path.exists())

            self.assertEqual(data, Path(path, "data").read_bytes())
            meta = json.loads(Path(path, "meta").read_text())
            self.assertIn("timestamp", meta)
            self.assertIn("shasum", meta)
            self.assertEqual(meta["labels"]["filename"], upload_path.name)


def rabbitmq_available():
    try:
        subprocess.check_output(["docker-compose", "exec", "rabbitmq", "true"])
        return True
    except subprocess.CalledProcessError:
        return False


def get_admin_connection():
    params = pika.ConnectionParameters(
        credentials=pika.PlainCredentials("admin", "admin")
    )
    return pika.BlockingConnection(params)


@unittest.skipUnless(rabbitmq_available(), "rabbitmq not available")
class TestPluginWithRabbitMQ(unittest.TestCase):
    def setUp(self):
        os.environ["WAGGLE_PLUGIN_USERNAME"] = "plugin"
        os.environ["WAGGLE_PLUGIN_PASSWORD"] = "plugin"
        os.environ["WAGGLE_PLUGIN_HOST"] = "127.0.0.1"
        os.environ["WAGGLE_PLUGIN_PORT"] = "5672"

        with get_admin_connection() as conn, conn.channel() as ch:
            ch.queue_purge("to-validator")

    def test_publish(self):
        now = time.time_ns()

        with Plugin() as publisher:
            publisher.publish("test", 123, meta={"sensor": "bme680"}, timestamp=now)

        with get_admin_connection() as conn, conn.channel() as ch:
            _, _, body = ch.basic_get("to-validator", auto_ack=True)
            msg = wagglemsg.load(body)

        self.assertEqual(
            msg,
            wagglemsg.Message(
                name="test",
                value=123,
                meta={"sensor": "bme680"},
                timestamp=now,
            ),
        )

    def test_subscribe(self):
        msg = wagglemsg.Message(
            name="test",
            value=123,
            meta={"sensor": "bme680"},
            timestamp=time.time_ns(),
        )

        with Plugin() as subscriber:
            subscriber.subscribe("test")
            time.sleep(1)

            with get_admin_connection() as conn, conn.channel() as ch:
                ch.basic_publish("data.topic", "test", wagglemsg.dump(msg))

            msg2 = subscriber.get(1)
            self.assertEqual(msg, msg2)


class TestPluginLogDir(unittest.TestCase):
    def test_log_dir(self):
        import sage_data_client

        with TemporaryDirectory() as dir:
            dir = Path(dir)

            # create dummy upload file
            upload_file = Path(dir, "hello.txt")
            upload_file.write_text("hello")

            timestamp = get_timestamp()

            # set env var and run plugin
            try:
                os.environ["PYWAGGLE_LOG_DIR"] = str(dir)
                with Plugin() as plugin:
                    plugin.publish("test", 123, timestamp=timestamp)
                    plugin.publish(
                        "test.with.meta",
                        456,
                        meta={"user": "data"},
                        timestamp=timestamp + 10000,
                    )
                    plugin.upload_file(upload_file, timestamp=timestamp + 20000)
            finally:
                del os.environ["PYWAGGLE_LOG_DIR"]

            df = sage_data_client.load(Path(dir, "data.ndjson"))

            # ensure records match what was published
            self.assertEqual(len(df), 3)

            # TODO(sean) test timestamps
            self.assertEqual(df.loc[0, "name"], "test")
            self.assertEqual(df.loc[0, "value"], 123)

            self.assertEqual(df.loc[1, "name"], "test.with.meta")
            self.assertEqual(df.loc[1, "value"], 456)
            self.assertEqual(df.loc[1, "meta.user"], "data")

            self.assertEqual(df.loc[2, "name"], "upload")
            self.assertEqual(df.loc[2, "meta.filename"], "hello.txt")

            # ensure all uploads exist
            for path in df[df.name == "upload"].value:
                self.assertTrue(Path(path).exists())


def assertDictContainsSubset(t, a, b):
    t.assertLessEqual(a.items(), b.items())


if __name__ == "__main__":
    unittest.main()



import unittest
from waggle.data.audio import AudioFolder, AudioSample
from waggle.data.vision import RGB, BGR, ImageFolder, ImageSample, resolve_device
from waggle.data.timestamp import get_timestamp
import numpy as np
from tempfile import TemporaryDirectory
from pathlib import Path
import os.path
from itertools import product


def generate_audio_data(samplerate, channels, dtype):
    if dtype == np.float32:
        return np.random.uniform(-1, 1, (samplerate, channels)).astype(dtype)
    if dtype == np.float64:
        return np.random.uniform(-1, 1, (samplerate, channels)).astype(dtype)
    if dtype == np.int16:
        return np.random.randint(
            -(2**15), 2**15, (samplerate, channels), dtype=dtype
        )
    if dtype == np.int32:
        return np.random.randint(
            -(2**31), 2**31, (samplerate, channels), dtype=dtype
        )
    raise ValueError("unsupported audio settings")


def generate_audio_sample(samplerate, channels, dtype):
    return AudioSample(generate_audio_data(samplerate, channels, dtype), samplerate, 0)


class TestData(unittest.TestCase):
    def test_colors(self):
        for fmt in [RGB, BGR]:
            data = np.random.randint(0, 255, (100, 120, 3), dtype=np.uint8)
            data2 = fmt.format_to_cv2(fmt.cv2_to_format(data))
            self.assertTrue(
                np.all(np.isclose(data, data2, 1.0)), f"checking format {fmt}"
            )

    def test_resolve_device(self):
        self.assertEqual(
            resolve_device(Path("test.jpg")), str(Path("test.jpg").absolute())
        )
        self.assertEqual(
            resolve_device("file://path/to/test.jpg"),
            str(Path("path/to/test.jpg").absolute()),
        )
        self.assertEqual(
            resolve_device("http://camera-ip.org/image.jpg"),
            "http://camera-ip.org/image.jpg",
        )
        self.assertEqual(
            resolve_device("rtsp://camera-ip.org/image.jpg"),
            "rtsp://camera-ip.org/image.jpg",
        )
        self.assertEqual(resolve_device(0), 0)

    def test_image_save(self):
        with TemporaryDirectory() as dir:
            sample = ImageSample(
                np.random.randint(0, 255, (100, 120, 3), dtype=np.uint8), 0, RGB
            )
            for fmt in ["jpg", "png"]:
                name = f"sample.{fmt}"
                sample.save(os.path.join(dir, name))
                sample.save(Path(dir, name))

    def test_image_save_load(self):
        with TemporaryDirectory() as dir:
            sample = ImageSample(
                np.random.randint(0, 255, (100, 120, 3), dtype=np.uint8), 0, RGB
            )
            sample.save(Path(dir, "sample.png"))
            samples = ImageFolder(dir, RGB)
            self.assertTrue(np.allclose(sample.data, samples[0].data))

    def test_audio_save(self):
        test_formats = ["wav", "flac"]
        test_samplerates = [22050, 44100, 48000]
        test_channels = [1, 2]
        test_dtypes = [np.float32, np.float64, np.int16, np.int32]

        for format, samplerate, channels, dtype in product(
            test_formats, test_samplerates, test_channels, test_dtypes
        ):
            name = f"sample.{format}"
            with TemporaryDirectory() as dir:
                sample = generate_audio_sample(
                    samplerate, channels=channels, dtype=dtype
                )
                # test saving as any PathLike
                sample.save(os.path.join(dir, name))
                sample.save(Path(dir, name))

    def test_audio_save_load(self):
        test_formats = ["wav", "flac"]
        test_samplerates = [22050, 44100, 48000]
        test_channels = [1, 2]
        test_dtypes = [np.float32, np.float64]

        for format, samplerate, channels, dtype in product(
            test_formats, test_samplerates, test_channels, test_dtypes
        ):
            name = f"sample.{format}"
            with TemporaryDirectory() as dir:
                sample = generate_audio_sample(
                    samplerate, channels=channels, dtype=dtype
                )
                sample.save(Path(dir, name))
                samples = AudioFolder(dir)
                self.assertTrue(
                    np.allclose(sample.data, samples[0].data, atol=1e-4),
                    msg=f"failed: format={format} samplerate={samplerate} channels={channels} dtypes={dtype}",
                )

    def test_get_timestamp(self):
        ts = get_timestamp()
        self.assertIsInstance(ts, int)


if __name__ == "__main__":
    unittest.main()



from datetime import datetime
import json
import time


class MeasurementsFile:
    def __init__(self, filename):
        self.records = []

        with open(filename, "r") as f:
            for r in map(json.loads, f):
                # 2021-06-25T18:52:15.404690128Z
                r["timestamp"] = datetime.strptime(
                    r["timestamp"][:26], "%Y-%m-%dT%H:%M:%S.%f"
                )
                self.records.append(r)
        self.records.sort(key=lambda r: r["timestamp"])

    def play(self, nodelay=False):
        if len(self.records) == 0:
            return
        last_record = self.records[0]
        for r in self.records:
            delta = r["timestamp"] - last_record["timestamp"]
            if not nodelay:
                time.sleep(delta.total_seconds())
            yield r
            last_record = r


# MessagePlayer can take a SDR format file and replay the contents
# this will help support use cases where someone wants to inject known
# data into their plugin from a file.
# (think about name?)
# other features:
# should sort by timestamp / or leave no sort as flag?
# should be able to decide starting time



from os import PathLike
from pathlib import Path
import numpy
import soundfile
from typing import NamedTuple
import random
from base64 import b64encode
from io import BytesIO
from .timestamp import get_timestamp


class AudioSample(NamedTuple):
    data: numpy.ndarray
    samplerate: int
    timestamp: int

    def save(self, path: PathLike):
        path = Path(path)
        soundfile.write(str(path), self.data, self.samplerate)

    def _repr_html_(self):
        with BytesIO() as buf:
            soundfile.write(
                buf, self.data, self.samplerate, format="flac", closefd=False
            )
            b64data = b64encode(buf.getvalue()).decode()
        return f"""
<audio controls="controls" autobuffer="autobuffer">
<source src="data:audio/wav;base64,{b64data}" />
</audio>
"""


class Microphone:
    def __init__(self, samplerate=48000, channels=1, name=None):
        import soundcard

        self.microphone = soundcard.default_microphone()
        self.samplerate = samplerate
        self.channels = channels
        self.name = name

    def record(self, duration):
        timestamp = get_timestamp()
        data = self.microphone.record(
            samplerate=self.samplerate,
            numframes=int(duration * self.samplerate),
            channels=self.channels,
        )
        return AudioSample(data, self.samplerate, timestamp=timestamp)


class AudioFolder:
    available_formats = {"." + s.lower() for s in soundfile.available_formats().keys()}

    def __init__(self, root, shuffle=False):
        self.files = sorted(
            p.absolute()
            for p in Path(root).glob("*")
            if p.suffix in self.available_formats
        )
        if shuffle:
            random.shuffle(self.files)

    def __len__(self):
        return len(self.files)

    def __getitem__(self, i):
        data, samplerate = soundfile.read(str(self.files[i]), always_2d=True)
        timestamp = Path(self.files[i]).stat().st_mtime_ns
        return AudioSample(data, samplerate, timestamp=timestamp)

    def __repr__(self):
        return f"AudioFolder{self.files!r}"



import cv2
from pathlib import Path
import numpy
from typing import Union
import os
from os import PathLike
import random
import json
import re
from base64 import b64encode
from .timestamp import get_timestamp


class BGR:
    @classmethod
    def cv2_to_format(cls, data):
        return data

    @classmethod
    def format_to_cv2(cls, data):
        return data


class RGB:
    @classmethod
    def cv2_to_format(cls, data):
        return cv2.cvtColor(data, cv2.COLOR_BGR2RGB)

    @classmethod
    def format_to_cv2(cls, data):
        return cv2.cvtColor(data, cv2.COLOR_RGB2BGR)


WAGGLE_DATA_CONFIG_PATH = Path(
    os.environ.get("WAGGLE_DATA_CONFIG_PATH", "/run/waggle/data-config.json")
)


def read_device_config(path):
    config = json.loads(Path(path).read_text())
    return {
        section["match"]["id"]: section
        for section in config
        if "id" in section["match"]
    }


# TODO use format spec like rgb vs bgr in config file
class ImageSample:
    data: numpy.ndarray
    timestamp: int
    format: Union[BGR, RGB]

    def __init__(self, data, timestamp, format):
        self.format = format
        self.data = self.format.cv2_to_format(data)
        self.timestamp = timestamp

    def save(self, path: PathLike):
        path = Path(path)
        data = self.format.format_to_cv2(self.data)
        cv2.imwrite(str(path), data)

    def _repr_html_(self):
        data = self.format.format_to_cv2(self.data)
        ok, buf = cv2.imencode(".png", data)
        if not ok:
            raise RuntimeError("could not encode image")
        b64data = b64encode(buf.ravel()).decode()
        return f'<img src="data:image/png;base64,{b64data}" />'


def resolve_device(device):
    if isinstance(device, Path):
        return resolve_device_from_path(device)
    # objects that are not paths or strings are considered already resolved
    if not isinstance(device, str):
        return device
    match = re.match(r"([A-Za-z0-9]+)://(.*)$", device)
    # non-url like paths refer to data shim devices
    if match is None:
        return resolve_device_from_data_config(device)
    # return file:// urls as path
    if match.group(1) == "file":
        return resolve_device_from_path(Path(match.group(2)))
    # return other urls as-is
    return device


def resolve_device_from_path(path):
    return str(path.absolute())


def resolve_device_from_data_config(device):
    config = read_device_config(WAGGLE_DATA_CONFIG_PATH)
    section = config.get(device)
    if section is None:
        raise KeyError(f"no device found {device!r}")
    try:
        return section["handler"]["args"]["url"]
    except KeyError:
        raise KeyError(f"missing .handler.args.url field for device {device!r}.")


class Camera:
    def __init__(self, device=0, format=RGB):
        self.capture = _Capture(resolve_device(device), format)

    def __enter__(self):
        self.capture.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.capture.__exit__(exc_type, exc_val, exc_tb)

    def snapshot(self):
        with self.capture:
            return self.capture.snapshot()

    def stream(self):
        with self.capture:
            yield from self.capture.stream()


class _Capture:
    def __init__(self, device, format):
        self.device = device
        self.format = format
        self.context_depth = 0

    def __enter__(self):
        if self.context_depth == 0:
            self.capture = cv2.VideoCapture(self.device)
            if not self.capture.isOpened():
                raise RuntimeError(
                    f"unable to open video capture for device {self.device!r}"
                )
        self.context_depth += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.context_depth -= 1
        if self.context_depth == 0:
            self.capture.release()

    def snapshot(self):
        timestamp = get_timestamp()
        ok, data = self.capture.read()
        if not ok:
            raise RuntimeError("failed to take snapshot")
        return ImageSample(data=data, timestamp=timestamp, format=self.format)

    def stream(self):
        while True:
            timestamp = get_timestamp()
            ok, data = self.capture.read()
            if not ok:
                break
            yield ImageSample(data=data, timestamp=timestamp, format=self.format)


class ImageFolder:
    available_formats = {".jpg", ".jpeg", ".png"}

    def __init__(self, root, format=RGB, shuffle=False):
        self.files = sorted(
            p.absolute()
            for p in Path(root).glob("*")
            if p.suffix in self.available_formats
        )
        self.format = format
        if shuffle:
            random.shuffle(self.files)

    def __len__(self):
        return len(self.files)

    def __getitem__(self, i):
        data = cv2.imread(str(self.files[i]))
        timestamp = Path(self.files[i]).stat().st_mtime_ns
        return ImageSample(data=data, timestamp=timestamp, format=self.format)

    def __repr__(self):
        return f"ImageFolder{self.files!r}"



from typing import NamedTuple


class PluginConfig(NamedTuple):
    """
    PluginConfig represents the config required to setup and run a Plugin.
    """

    # TODO generalize to support different backends
    username: str
    password: str
    host: str
    port: int
    app_id: str



import time

# BUG This *must* be addressed with the behavior written up in the plugin spec.
# We don't want any surprises in terms of accuraccy
try:
    from time import time_ns as get_timestamp
except ImportError:

    def get_timestamp():
        return int(time.time() * 1e9)


# NOTE to preserve the best accuracy, we implement the backwards compatible perf
# counter by only abstracting how to measure the duration between two times in
# nanoseconds
try:
    from time import perf_counter_ns as timeit_perf_counter

    def timeit_perf_counter_duration(start, finish):
        return finish - start

except ImportError:
    from time import perf_counter as timeit_perf_counter

    def timeit_perf_counter_duration(start, finish):
        return int((finish - start) * 1e9)



import logging
import numpy as np
from urllib.request import urlopen
from threading import Thread, Event
from queue import Queue, Empty, Full
import time
import os
import socket
from pathlib import Path
import json
import random
import re

logger = logging.getLogger(__name__)

try:
    import cv2
except ImportError:
    logger.warning(
        "cv2 module not found. pywaggle requires this to capture image and video data."
    )


# BUG This *must* be addressed with the behavior written up in the plugin spec.
# We don't want any surprises in terms of accuraccy
try:
    from time import time_ns
except ImportError:
    logger.warning("using backwards compatible implementation of time_ns")

    def time_ns():
        return int(time.time() * 1e9)


def cvtColor(bgr_img, pixel_format="rgb"):
    if pixel_format == "rgb":
        return cv2.cvtColor(bgr_img, cv2.COLOR_BGR2RGB)
    return bgr_img


class ImageHandler:
    def __init__(self, query, url, pixel_format="rgb"):
        self.url = url
        self.pixel_format = pixel_format

    def get(self, timeout=None):
        try:
            with urlopen(self.url, timeout=timeout) as f:
                data = f.read()
                ts = time_ns()
                arr = np.frombuffer(data, np.uint8)
                bgr_img = cv2.imdecode(arr, cv2.IMREAD_COLOR)
                return ts, cvtColor(bgr_img, self.pixel_format)
        except socket.timeout:
            raise TimeoutError("get timed out")

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass


def video_worker(handler):
    try:
        while not handler.quit.is_set():
            ok, bgr_img = handler.cap.read()
            if not ok:
                break
            img = cvtColor(bgr_img, handler.pixel_format)
            item = (time_ns(), img)

            # attempt to add an item to the queue
            try:
                handler.queue.put_nowait(item)
                continue
            except Full:
                logger.debug("video frame queue full. evicting oldest frame...")
            # evict an item from queue
            try:
                handler.queue.get_nowait()
            except Empty:
                pass
            # queue should have space to add now. (assuming this
            # is the only producer adding to this queue)
            handler.queue.put_nowait(item)
    finally:
        handler.cap.release()
        handler.released.set()


# TODO We need to use a flexible model where the data returned is
# extensible. For example, serial data won't really have a good
# notion of "timestamp". Maybe it's better to not include that.


class VideoHandler:
    def __init__(self, query, url, pixel_format="rgb"):
        self.pixel_format = pixel_format
        self.cap = cv2.VideoCapture(url)
        if not self.cap.isOpened():
            raise RuntimeError(f'could not open camera at "{url}".')
        self.queue = Queue(8)
        self.quit = Event()
        self.released = Event()
        # NOTE(sean) no further mutation can be done on VideoHandler state. all
        # interaction with cap *must* be done in the worker thread or via queue
        # and quit primitives
        worker = Thread(target=video_worker, args=(self,), daemon=True)
        worker.start()

    def get(self, timeout=None):
        try:
            return self.queue.get(timeout=timeout)
        except Empty:
            raise TimeoutError("get timed out")

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.quit.set()
        self.released.wait()  # <- wait for cleanup in worker thread


WAGGLE_DATA_CONFIG_PATH = Path(
    os.environ.get("WAGGLE_DATA_CONFIG_PATH", "/run/waggle/data-config.json")
)

try:
    config = json.loads(WAGGLE_DATA_CONFIG_PATH.read_text())
except FileNotFoundError:
    logger.debug(
        "could not find data config file %s. using empty resource list.",
        WAGGLE_DATA_CONFIG_PATH,
    )
    config = []


def dict_is_subset(a, b):
    return all(k in b and re.match(b[k], a[k]) for k in a.keys())


def find_all_matches(query):
    return [c for c in config if dict_is_subset(query, c["match"])]


def find_match(query):
    matches = find_all_matches(query)
    if len(matches) == 0:
        raise RuntimeError("no matches found")
    if len(matches) > 1:
        raise RuntimeError("multiple devices found")
    return matches[0]


handlers = {
    "image": ImageHandler,
    "video": VideoHandler,
}


# optimizations *could* happen here, on demand...
def open_data_source(**query):
    match = find_match(query)
    handler = handlers[match["handler"]["type"]]
    args = match["handler"]["args"]
    return handler(query, **args)



import logging
import re
import wagglemsg

from contextlib import contextmanager
from datetime import datetime
from os import getenv
from pathlib import Path
from queue import Queue, Empty
from threading import Event
from typing import NamedTuple

from .config import PluginConfig
from .rabbitmq import RabbitMQPublisher, RabbitMQConsumer
from .time import get_timestamp, timeit_perf_counter, timeit_perf_counter_duration
from .uploader import Uploader


logger = logging.getLogger(__name__)


class PublishData(NamedTuple):
    scope: str
    body: bytes


# Nanoseconds since epoch for 2000-01-01T00:00:00Z
MIN_TIMESTAMP_NS = 946706400000000000


class FilesystemPublisher:
    def __init__(self, root):
        self.root = Path(root)
        self.root.mkdir(parents=True, exist_ok=True)
        self.datafile = Path(root, "data.ndjson").open("a")
        self.uploads_dir = Path(root, "uploads")
        self.uploads_dir.mkdir(parents=True, exist_ok=True)

    def close(self):
        self.datafile.close()

    def publish(self, msg: wagglemsg.Message):
        import json

        out = {
            "name": msg.name,
            "value": msg.value,
            "meta": msg.meta,
            # python doesn't have builtin support for nanosecond
            "timestamp": isoformat_time_ns(msg.timestamp),
        }
        print(
            json.dumps(out, sort_keys=True, separators=(",", ":")),
            file=self.datafile,
            flush=True,
        )

    def upload_file(self, path, timestamp, meta):
        from shutil import copyfile

        src = Path(path)
        dst = Path(self.uploads_dir, f"{timestamp}-{src.name}")
        copyfile(src, dst)
        meta = meta.copy()
        meta["filename"] = Path(src).name
        self.publish(
            wagglemsg.Message(
                name="upload",
                value=str(dst.absolute()),
                meta=meta,
                timestamp=timestamp,
            )
        )


def isoformat_time_ns(ns: int) -> str:
    # python doesn't have builtin support for nanosecond timestamps and formatting, so we provide
    # a backfill for it. this is only intended to be used in the run log for testing.
    nanostr = f"{ns%1000:03d}"
    return datetime.fromtimestamp(ns / 1e9).isoformat() + nanostr


class Plugin:
    """
    Plugin provides methods to publish and consume messages inside the Waggle ecosystem.

    Examples
    --------

    The simplest example is creating a Plugin and publishing a message. This can be done using:

    ```python
    from waggle.plugin import Plugin

    with Plugin() as plugin:
        plugin.publish("test_value", 99)
    ```
    """

    def __init__(
        self, config=None, uploader=None, file_publisher: FilesystemPublisher = None
    ):
        self.config = config or get_default_plugin_config()
        self.uploader = uploader or get_default_plugin_uploader()
        self.send = Queue()
        self.recv = Queue()
        self.stop = Event()
        self.tasks = []

        # TODO(sean) can we use ExitStack to clean up???

        self.file_publisher = file_publisher

        if self.file_publisher is None and getenv("PYWAGGLE_LOG_DIR") is not None:
            self.file_publisher = FilesystemPublisher(getenv("PYWAGGLE_LOG_DIR"))

    def __enter__(self):
        self.tasks.append(RabbitMQPublisher(self.config, self.send, self.stop))
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.stop.set()

        if self.file_publisher is not None:
            self.file_publisher.close()

        for task in self.tasks:
            task.done.wait()

    def subscribe(self, *topics):
        self.tasks.append(RabbitMQConsumer(topics, self.config, self.recv, self.stop))
        # TODO(sean) add mock or integration testing against rabbitmq to actually test this

    def get(self, timeout=None):
        try:
            return self.recv.get(timeout=timeout)
        except Empty:
            pass
        raise TimeoutError("plugin get timed out")

    def publish(self, name, value, meta={}, timestamp=None, scope="all", timeout=None):
        # get timestamp before doing other work
        timestamp = timestamp or get_timestamp()
        raise_for_invalid_publish_name(name)
        self.__publish(name, value, meta, timestamp, scope, timeout)

    # NOTE __publish is used internally by publish and upload_file to do an unchecked
    # message publish. the main reason this exists is to guard against reserved names
    # like "upload" in publish but still allow upload_file to use it.
    def __publish(self, name, value, meta, timestamp, scope="all", timeout=None):
        if not isinstance(value, (int, float, str)):
            raise TypeError("Value must be an int, float or str.")
        if not isinstance(timestamp, int):
            raise TypeError(
                "Timestamp must be an int and have units of nanoseconds since epoch. Please see the documentation for more information on setting timestamps."
            )
        if timestamp < MIN_TIMESTAMP_NS:
            raise ValueError(
                "Timestamp probably has wrong units and is being processed as before 2000-01-01T00:00:00Z. Timestamp must have units of nanoseconds since epoch. Please see the documentation for more information on setting timestamps."
            )
        if not valid_meta(meta):
            raise TypeError("Meta must be a dictionary of strings to strings.")
        msg = wagglemsg.Message(name=name, value=value, timestamp=timestamp, meta=meta)

        # hack to use file publisher for everything except uploads
        if self.file_publisher is not None and name != "upload":
            self.file_publisher.publish(msg)

        logger.debug("adding message to outgoing queue: %s", msg)
        self.send.put(PublishData(scope, wagglemsg.dump(msg)), timeout=timeout)

    def upload_file(self, path, meta={}, timestamp=None, keep=False):
        # get timestamp before doing other work
        timestamp = timestamp or get_timestamp()

        if self.file_publisher is not None:
            self.file_publisher.upload_file(path, meta=meta, timestamp=timestamp)

        if self.uploader is not None:
            meta = meta.copy()
            meta["filename"] = Path(path).name
            upload_path = self.uploader.upload_file(
                path=path, meta=meta, timestamp=timestamp, keep=keep
            )
            self.__publish("upload", upload_path.name, meta, timestamp)

    @contextmanager
    def timeit(self, name):
        logger.debug("starting timeit block %s", name)
        start = timeit_perf_counter()
        yield
        finish = timeit_perf_counter()
        duration = timeit_perf_counter_duration(start, finish)
        self.publish(name, duration)
        logger.debug("finished timeit block %s", name)


def get_default_plugin_config() -> PluginConfig:
    return PluginConfig(
        username=getenv("WAGGLE_PLUGIN_USERNAME", "plugin"),
        password=getenv("WAGGLE_PLUGIN_PASSWORD", "plugin"),
        host=getenv("WAGGLE_PLUGIN_HOST", "rabbitmq"),
        port=int(getenv("WAGGLE_PLUGIN_PORT", 5672)),
        app_id=getenv("WAGGLE_APP_ID", ""),
    )


def valid_meta(meta):
    return isinstance(meta, dict) and all(isinstance(v, str) for v in meta.values())


def get_default_plugin_uploader():
    if (
        getenv("WAGGLE_PLUGIN_UPLOAD_PATH") is None
        and getenv("PYWAGGLE_LOG_DIR") is not None
    ):
        return None
    return Uploader(Path(getenv("WAGGLE_PLUGIN_UPLOAD_PATH", "/run/waggle/uploads")))


publish_name_part_pattern = re.compile("^[a-z0-9_]+$")


def raise_for_invalid_publish_name(s: str):
    if not isinstance(s, str):
        raise TypeError(f"publish name must be a string: {s!r}")
    if len(s) > 128:
        raise ValueError(f"publish must be at most 128 characters: {s!r}")
    if s == "upload":
        raise ValueError(f"name {s!r} is reserved for system use only")
    parts = s.split(".")
    for p in parts:
        if not publish_name_part_pattern.match(p):
            raise ValueError(
                f"publish name invalid: {s!r} part: {p!r} (names must consist of [a-z0-9_] and may be joined by .)"
            )



import hashlib
import json
from pathlib import Path
from shutil import copyfile
from .time import get_timestamp


class Uploader:
    def __init__(self, root):
        self.root = Path(root)

    # NOTE uploads are stored in the following directory structure:
    # root/
    #   timestamp-sha1sum/
    #     data
    #     meta
    def upload_file(self, path, meta={}, timestamp=None, keep=False):
        # get timestamp before doing other work
        timestamp = timestamp or get_timestamp()

        path = Path(path)
        checksum = sha1sum_for_file(path)

        # create upload dir
        upload_dir = Path(self.root, f"{timestamp}-{checksum}")
        upload_dir.mkdir(parents=True, exist_ok=True)

        # stage data file
        # NOTE we do a copy instead of move, as the upload dir may
        # be mounted from another disk.
        copyfile(path, Path(upload_dir, "data"))
        if not keep:
            path.unlink()

        # stage meta file
        metafile = {
            "timestamp": timestamp,
            "shasum": checksum,
            "labels": {k: v for k, v in meta.items()},
        }
        metafile["labels"]["filename"] = path.name
        write_json_file(Path(upload_dir, "meta"), metafile)

        return upload_dir


def sha1sum_for_file(path):
    h = hashlib.sha1()
    with open(path, "rb") as f:
        while True:
            chunk = f.read(32768)
            if chunk == b"":
                break
            h.update(chunk)
    return h.hexdigest()


def write_json_file(path, obj):
    with open(path, "w") as f:
        json.dump(obj, f, separators=(",", ":"), sort_keys=True)
